package gnu.brl;

// BRLRead.java -- a class to read BRL forms
// Copyright (C) 1999, 2000  Bruce R. Lewis and Eaton Vance Management
// See the file COPYING for license terms.

import java.io.*;
import kawa.lang.*;
import gnu.text.SyntaxException;
import gnu.mapping.*;
import gnu.expr.*;
import gnu.kawa.util.*;
import gnu.text.SourceMessages;

public class BRLRead extends kawa.lang.ScmRead
{
    public BRLRead(InPort port)
    {
	super(port);
    }

    public BRLRead(InPort port, SourceMessages messages)
    {
	super(port, messages);
    }

    protected boolean isDelimiter (char ch)
    {
	return (Character.isWhitespace (ch)
		|| ch == ')' || ch == '(' || ch == '"' || ch == ';'
		|| ch == '[' || ch == ']');
    }

    protected int skipWhitespaceAndComments()
	throws java.io.IOException
    {
	if (port.getLineNumber() <= 1
	    && port.getColumnNumber() <= 1)
	    // At the beginning of a BRL file don't skip
	    return read();
		
	else
	    // Elsewhere skip as in normal Scheme
	    return super.skipWhitespaceAndComments();
    }

    /**
     * Read a BRL string literal.
     * Assume we have already skipped the initial ']'
     * or that we're at the beginning of the file.
     */
    protected Object readBRLLiteral()
	throws java.io.IOException, SyntaxException
    {
	StringBuffer obj = new StringBuffer ();
	boolean inString = true;
	int c = ']';
	int prev;
	do
	    {
		int next, v;

		prev = c;
		if (port.pos < port.limit
		    && prev != '\r' && prev != '\n')
		    c = port.buffer[port.pos++];
		else
		    c = read();

		switch (c)
		    {
		    case '[':
			if (peek() == '[')
			    {
				obj.append((char) c);
				skip();
			    }
			else
			    {
				inString = false;
			    }
			break;
		    default:
			if (c < 0)
			    inString = false;
			else
			    obj.append ((char) c);
			break;
		    }
	    } while (inString);
	return new FString (obj);
    }

    public static Object readObject(InPort port)
	throws java.io.IOException, SyntaxException
    {
	return (new BRLRead(port)).readObject();
    }

    public Object readObject (int c)
	throws java.io.IOException, SyntaxException
    {
	if (port.getLineNumber() <= 1
	    && port.getColumnNumber() <= 2)
	    {
		// At the beginning of file, read a literal w/o starting ']'
		unread(c);
		return readBRLLiteral();
	    }
	else
	    // otherwise read normally
	    return super.readObject(c);
    }

    public Object readDefault (int c)
	throws java.io.IOException, SyntaxException
    {
	if (c == ']')
	    return readBRLLiteral();
	else
	    return super.readDefault(c);
    }
}
