;;; brl-global.scm -- global definitions for the BRL servlet
;;; Copyright (C) 1999, 2000  Bruce R. Lewis and Eaton Vance Management
;;; See the file COPYING for license terms.

;;;;; Part I: NAL
;;;
;;; nal - named-argument lambda
;;; A nal is a list whose first element is a procedure,
;;; and whose remaining elements are symbols naming the
;;; second and subsequent arguments to the procedure.
;;; The first argument is brl-context, explained later.
;;;

(define brl-nal-proc car)
(define brl-nal-args cdr)

;
;  Take a list of expressions (including the inputs expr)
;  and create a nal from it.
;

(define (brl-make-nal l)
  (let ((ins (brl-template-inputs l)))
    (cons (eval `(lambda ,(cons 'brl-context ins)
		   (letrec-syntax
		       ((brl-out
			 (syntax-rules (brl-set-cont!
					brl-set-endproc!
					brl-set-outport!
					define define-syntax
					set! set-car! set-cdr!
					sql-statement-close
					string-fill! string-set!
					vector-fill! vector-set!
					write)
			   ((brl-out (brl-set-cont! expr ...))
			    (brl-set-cont! expr ...))
			   ((brl-out (brl-set-endproc! expr ...))
			    (brl-set-endproc! expr ...))
			   ((brl-out (brl-set-outport! expr ...))
			    (brl-set-outport! expr ...))
			   ((brl-out (define expr ...)) (define expr ...))
			   ((brl-out (define-syntax expr ...))
			    (define-syntax expr ...))
			   ((brl-out (set! expr ...)) (set! expr ...))
			   ((brl-out (set! expr ...)) (set! expr ...))
			   ((brl-out (set-car! expr ...)) (set-car! expr ...))
			   ((brl-out (set-cdr! expr ...)) (set-cdr! expr ...))
			   ((brl-out (sql-statement-close expr ...)) (sql-statement-close expr ...))
			   ((brl-out (string-fill! expr ...)) (string-fill! expr ...))
			   ((brl-out (string-set! expr ...)) (string-set! expr ...))
			   ((brl-out (vector-fill! expr ...)) (vector-fill! expr ...))
			   ((brl-out (vector-set! expr ...)) (vector-set! expr ...))
			   ((brl-out (write expr ...)) (write expr ...))
			   ((brl-out expr) (display expr (brl-context-outport
						      brl-context)))
			   ((brl-out expr1 expr2 ...) (begin
						    (brl-out expr1)
						    (brl-out expr2 ...)))))
			(brl
			 (syntax-rules ()
			   ((brl expr ...)
			    (begin (brl-out expr ...) ""))))
			(cgi
			 (syntax-rules ()
			   ((cgi var) ((; constant-fold ;FIXME
					brl-sv-req-string-retriever
					(quote var))
				       (brl-context-sv-req brl-context))))))
		     ,(cons 'brl (brl-template-inputs-delete! l l)))))
	  ins)))

;;; brl-context - a list with at least two elements

(define brl-context-outport car)	; an output port
(define brl-context-ctin cadr)		; a continuation
(define brl-context-endproc caddr)	; a procedure or #f
; optional elements
(define brl-context-bindings cadddr)	; list of input symbols and values
(define (brl-context-type c) (list-ref c 4)); a symbol - currently just 'servlet
(define (brl-context-sv-req c) (list-ref c 5)) ; Servlet HTTP request
(define (brl-context-sv-rsp c) (list-ref c 6)) ; Servlet HTTP response

; Cookies
(define (brl-context-cookies c)
  ((primitive-static-method
    <gnu.brl.brlsv>
    "getSchemeCookies" <Object> (<javax.servlet.http.HttpServletRequest>))
   (brl-context-sv-req c)))
(define (brl-cookie-value context sym)
  (let ((triplet (assoc sym (brl-context-cookies context))))
    (if triplet
	(cadr triplet)
	#f)))

(define brl-set-outport! set-car!)
(define (brl-continue c) ((brl-context-ctin c)))
(define (brl-set-cont! c ctin) (set-car! (cdr c) ctin))
(define (brl-set-endproc! c proc) (set-car! (cddr c) proc))

; For getting strings out of an HttpServletRequest object
(define (brl-sv-req-string-retriever cgi-variable)
  (let ((method
	 (cond ((eq? cgi-variable 'SERVER_NAME)
		(primitive-interface-method <javax.servlet.ServletRequest>
					  "getServerName" <String> ()))
	       ((eq? cgi-variable 'SERVER_PORT)
		(primitive-interface-method <javax.servlet.ServletRequest>
					  "getServerPort" <String> ()))
	       ((eq? cgi-variable 'REMOTE_ADDR)
		(primitive-interface-method <javax.servlet.ServletRequest>
					  "getRemoteAddr" <String> ()))
	       ((eq? cgi-variable 'REMOTE_HOST)
		(primitive-interface-method <javax.servlet.ServletRequest>
					  "getRemoteHost" <String> ()))
	       ((eq? cgi-variable 'REQUEST_METHOD)
		(primitive-interface-method
		 <javax.servlet.http.HttpServletRequest>
		 "getMethod" <String> ()))
	       ((eq? cgi-variable 'PATH_INFO)
		(primitive-interface-method
		 <javax.servlet.http.HttpServletRequest>
		 "getPathInfo" <String> ()))
	       ((eq? cgi-variable 'PATH_TRANSLATED)
		(primitive-interface-method
		 <javax.servlet.http.HttpServletRequest>
		 "getPathTranslated" <String> ()))
	       ((eq? cgi-variable 'REMOTE_USER)
		(primitive-interface-method
		 <javax.servlet.http.HttpServletRequest>
		 "getRemoteUser" <String> ()))
	       ((eq? cgi-variable 'AUTH_TYPE)
		(primitive-interface-method
		 <javax.servlet.http.HttpServletRequest>
		 "getAuthType" <String> ()))
	       (else #f))))
    (if method
	(lambda (req) (kawa-convert (method req)))
	(let* ((cgi-str (brl-string cgi-variable))
	       (header-sym (and (> (string-length cgi-str) 5)
				(string-ci=? "HTTP_"
					     (substring
					      (brl-string cgi-variable)
					      0 5))
				(invoke (string->symbol
					 (substring cgi-str 5
						    (string-length cgi-str)))
					'replace #\_ #\-))))
	  (if header-sym
	      (let ((method2 
		     (primitive-interface-method
		      <javax.servlet.http.HttpServletRequest>
		      "getHeader" <String> (<String>))))
		(lambda (req)
		  (kawa-convert
		   (method2 req header-sym))))
	      (lambda (req) '()))))))

;;;;; Part II: functions to read BRL template files

;
; A BRL template may have an expr (inputs a b c ...) to ease compilation
; This is never the 1st exp in the template.
;

(define (brl-template-inputs l)
  (if (null? l)
      '()
      (if (and (list? (car l))
	       (eq? 'inputs (caar l)))
	  (cdar l)
	  (brl-template-inputs (cdr l)))))

(define (brl-template-inputs-delete! l start)
  (if (or (not (pair? l))
	  (not (pair? (cdr l))))
      start
      (if (and (pair? (cadr l))
	       (eq? 'inputs (caadr l)))
	  (begin
	    (set-cdr! l (cddr l))
	    start)
	  (brl-template-inputs-delete! (cdr l) start))))

(define (port-readall p accum)
  (let ((nextobj (read p)))
    (if (eof-object? nextobj)
	accum
	(if (null? accum)
	    (port-readall p (list nextobj))
	    (port-readall p (cons nextobj accum))))))

(define (brl-port-readall p)
  (brl-port-init p)
  (reverse (port-readall p '())))

(define (brl-read-nal p)
  (brl-make-nal (brl-readall p)))

(define (brl-load fname)
  (call-with-input-file fname brl-read-nal))

(define (brl-result nal args . partial-context)
  (let ((full-context (cons #f (cons #f (cons #f partial-context)))))
    (call-with-output-string
     (lambda (p)
       (try-finally
	(call-with-current-continuation
	 (lambda (ctin)
	   (brl-set-outport! full-context p)
	   (brl-set-cont! full-context ctin)
	   (apply (car nal) (cons full-context args))))
	(let ((endproc (brl-context-endproc full-context)))
	  (if endproc (endproc))))))))

;;;;; Part III: binding sets

(define (brl-binding-make name val) (cons name val))

(define (brl-binding-add bset b)
  (let ((existing (assq (car b) bset)))
    (if existing
	(begin
	  (set-cdr! existing
		    ((if (list? (cdr existing))
			 cons
			 list) (cdr b) (cdr existing)))
	  bset)
	(cons b bset))))

(define (brl-bindings bset blist)
  (if (null? blist)
      bset
      (brl-bindings (brl-binding-add bset (car blist))
		    (cdr blist))))

(define (brl-apply nal bset . partial-context)
  (apply
   brl-result
   (cons nal
	 (cons (map (lambda (sym)
		      (let ((binding (assq sym bset)))
			(if binding
			    (cdr binding)
			    '())))
		    (cdr nal))
	       partial-context))))

;
; Rudimentary cache of NALs.  It's a list, not a hash table,
; but should be o.k. for a few hundred NALs.
;

(define brl-nal-cache '())

(define (brl-nal-cache-find fname cache)
  (assoc fname cache))

(define (brl-handle-request fname blist . partial-context)
  (let ((cached (assoc fname brl-nal-cache))
	(new-modtime (file-last-modified fname)))
    (let ((nal
	   (if cached
	       (if (not (= new-modtime (caddr cached)))
		   (synchronized
		    cached
		    (begin (set-car! (cdr cached)
				     (brl-load fname))
			   (set-car! (cddr cached) new-modtime)
			   (cadr cached)))
		   (cadr cached))
	       (synchronized
		brl-nal-cache
		(set! brl-nal-cache
		      (cons (list fname (brl-load fname) new-modtime)
			    brl-nal-cache))
		(cadar brl-nal-cache)))))
      (apply brl-apply
	     (cons nal
		   (cons (brl-bindings '() blist)
			 (cons blist partial-context)))))))

(define (file-last-modified fname)
  (kawa-convert
   ((primitive-virtual-method <java.io.File>
			      "lastModified" <long> ())
    ((primitive-constructor <java.io.File> (<String>))
     fname))))

; Any object --> string

(define (brl-string obj)
  (cond ((string? obj) obj)
	((number? obj) (number->string obj))
	(else (call-with-output-string
	       (lambda (p) (display obj p))))))


;;;;; Part IV: JDBC procedures

(define (sql-driver classname)
  ((primitive-static-method
    <java.lang.Class> "forName" <java.lang.Class> (<String>))
   classname))

(define (sql-connection url . properties)
  (case (length properties)
    ((0) ((primitive-static-method <java.sql.DriverManager> "getConnection"
				   <java.sql.Connection> (<String>))
	  url))
    ((1) ((primitive-static-method <java.sql.DriverManager> "getConnection"
	   <java.sql.Connection> (<String> <java.util.Properties>))
	  url (car properties)))
    ((2) ((primitive-static-method <java.sql.DriverManager> "getConnection"
	   <java.sql.Connection> (<String> <String> <String>))
	  url (car properties) (cadr properties)))))

(define (sql-statement conn)
  ((primitive-interface-method
    <java.sql.Connection> "createStatement" <java.sql.Statement> ())
   conn))

(define (sql-statement-close statement)
  ((primitive-interface-method
    <java.sql.Statement> "close" <void> ())
   statement))

(define (sql-execute-update statement update)
  ((primitive-constructor <integer> (<int>))
   ((primitive-interface-method
     <java.sql.Statement> "executeUpdate" <int> (<String>))
    statement update)))

(define (sql-execute-query statement query)
  ((primitive-interface-method
    <java.sql.Statement> "executeQuery" <java.sql.ResultSet> (<String>))
   statement query))

(define (sql-rsmd resultset)
  ((primitive-interface-method
    <java.sql.ResultSet> "getMetaData" <java.sql.ResultSetMetaData> ())
   resultset))

(define (sql-rsmd-columncount rsmd)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnCount" <int> ())
   rsmd))

(define (sql-rsmd-columnlabel rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnLabel" <String> (<int>))
   rsmd i))

(define (sql-rsmd-columnname rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnName" <String> (<int>))
   rsmd i))

(define (sql-rsmd-columntypename rsmd i)
  ((primitive-interface-method
    <java.sql.ResultSetMetaData> "getColumnTypeName" <String> (<int>))
   rsmd i))

(define (sql-rsmd-iter op rsmd colnum accum)
  (if (zero? colnum)
      accum
      (sql-rsmd-iter
       op rsmd (- colnum 1) (cons (kawa-convert (op rsmd colnum)) accum))))

(define (sql-rsmd-all-columnlabels rsmd)
  (sql-rsmd-iter sql-rsmd-columnlabel rsmd (sql-rsmd-columncount rsmd) '()))

(define (sql-rsmd-all-columnnames rsmd)
  (sql-rsmd-iter sql-rsmd-columnname rsmd (sql-rsmd-columncount rsmd) '()))

(define (sql-rsmd-all-columntypenames rsmd)
  (sql-rsmd-iter sql-rsmd-columntypename rsmd (sql-rsmd-columncount rsmd) '()))

(define (sql-resultset-readnext resultset)
  ((primitive-interface-method
    <java.sql.ResultSet> "next" <boolean> ())
   resultset))

(define (sql-resultset-getobject resultset i)
  ((primitive-interface-method
    <java.sql.ResultSet> "getObject" <Object> (<int>))
   resultset i))

;; Convert a Java object, esp. one returned by JDBC, to
;; the appropriate Kawa object.

(define (kawa-convert java-object)
  (cond
   ((eq? java-object #!null) '())
   ((or (instance? java-object <java.lang.String>)
	(instance? java-object <java.lang.StringBuffer>))
    ((primitive-constructor <string> (<java.lang.String>))
     (as <java.lang.String> java-object)))
;; java.util.Date methods are mostly deprecated, but that class
;; more closely resembles the date type in 1990s SQL servers.
;; As more SQL servers handle timezones, GregorianCalendar will be better.
;; Code commented out for now so java.util.Date will stay java.util.Date
;   ((instance? java-object <java.util.Date>)
;    (let ((cal ((primitive-constructor <java.util.GregorianCalendar> ()))))
;      ((primitive-virtual-method
;	<java.util.Calendar> "setTime" <void> (<java.util.Date>))
;       cal java-object)
;      cal))
   ((or (instance? java-object <java.lang.Integer>)
	(instance? java-object <java.lang.Short>))
    ((primitive-constructor <integer> (<int>))
     (as <int> java-object)))
   ((instance? java-object <java.io.InputStream>)
    ((primitive-constructor <string> (<java.lang.String>))
     (invoke java-object 'toString)))
   (else java-object)))

; FIXME: other date-related functions would be useful.

(define (sql-resultset-consrow resultset colnum accum)
  (if (< colnum 1)
      accum
      (sql-resultset-consrow resultset (- colnum 1)
			     (cons (kawa-convert
				    (sql-resultset-getobject
				     resultset colnum))
				   accum))))

(define (sql-resultset-nextrow resultset colcount)
  (if (sql-resultset-readnext resultset)
      (sql-resultset-consrow resultset colcount '())
      #f))

(define (sql-resultset-consall resultset colcount accum)
  (if (sql-resultset-readnext resultset)
      (sql-resultset-consall resultset colcount
			     (cons (sql-resultset-consrow
				    resultset colcount '())
				   accum))
      accum))

(define (sql-statement-results stmt query)
  (let* ((rs (sql-execute-query stmt query))
	 (rsmd (sql-rsmd rs))
	 (colcount (sql-rsmd-columncount rsmd)))
    (reverse (sql-resultset-consall rs colcount '()))))

(define (sql-change? proc row1 row2)
  (or (null? row2)
      (not row2)
      (not (equal? (apply proc row1)
		   (apply proc row2)))))

(define-syntax sql-repeat-rsmd
  (syntax-rules (!body! new? last?)
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row (new? expr))
     (sql-change? (lambda args expr) sql-row prior-row))
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row (last? expr))
     (sql-change? (lambda args expr) sql-row next-row))

    ; Handle various scheme syntaxes
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (quote expr))
     (quote expr))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (define name expr ...))
     (define name (sql-repeat-rsmd !body! args prior-row sql-row next-row (begin expr ...))))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (set! name expr ...))
     (set! name (sql-repeat-rsmd !body! args prior-row sql-row next-row (begin expr ...))))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (lambda formals expr ...))
     (lambda formals (sql-repeat-rsmd !body! args prior-row sql-row next-row (begin expr ...))))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (let ((var init) ...) expr ...))
     (let ((var (sql-repeat-rsmd !body! args prior-row sql-row next-row init)) ...)
       (sql-repeat-rsmd !body! args prior-row sql-row next-row expr) ...))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (let* ((var init) ...) expr ...))
     (let* ((var (sql-repeat-rsmd !body! args prior-row sql-row next-row init)) ...)
       (sql-repeat-rsmd !body! args prior-row sql-row next-row expr) ...))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (letrec ((var init) ...) expr ...))
     (letrec ((var (sql-repeat-rsmd !body! args prior-row sql-row next-row init)) ...)
       (sql-repeat-rsmd !body! args prior-row sql-row next-row expr) ...))

    ;; case syntax -- FIXME: handle new? and last? in clauses
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row
	      (case key clause ...))
     (let ((tmp (sql-repeat-rsmd !body! args prior-row sql-row next-row key)))
       (case tmp
	 clause ...)))

    ((sql-repeat-rsmd !body! args prior-row sql-row next-row (brl expr ...))
     (brl
      (sql-repeat-rsmd !body! args prior-row sql-row next-row expr) ...))

    ;; Procedure application and other simple syntax
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row (expr ...))
     ((sql-repeat-rsmd !body! args prior-row sql-row next-row expr) ...))
    ((sql-repeat-rsmd !body! args prior-row sql-row next-row expr)
     expr)
    ((sql-repeat-rsmd rsmd statement args (query ...) expr ...)
     (let* ((sql-results
	     (sql-execute-query statement
				(apply string-append (map brl-string
							  (list query ...)))))
	    (rsmd (sql-rsmd sql-results))
	    (sql-columncount (sql-rsmd-columncount rsmd))
	    (first-row (sql-resultset-nextrow sql-results sql-columncount)))
       (do ((sql-row first-row next-row)
	    (next-row (if first-row
			  (sql-resultset-nextrow sql-results sql-columncount)
			  #f)
		      (if next-row
			  (sql-resultset-nextrow sql-results sql-columncount)
			  #f))
	    (prior-row #f sql-row)
	    (sql-rowcount 1 (+ sql-rowcount 1)))
	   ((not sql-row) (- sql-rowcount 1))
	 (apply (lambda args
		  (sql-repeat-rsmd !body! args prior-row sql-row next-row
				   (begin expr ...)))
		sql-row))))))

(define-syntax sql-repeat
  (syntax-rules ()
    ((sql-repeat statement args (query ...) expr ...)
     (sql-repeat-rsmd rsmd statement args (query ...) expr ...))))

;;
;; Text formatting
;;
(define (brl-SimpleDateFormat pattern)
  ((primitive-constructor <java.text.SimpleDateFormat>
			  (<String>))
   pattern))

(define (brl-SimpleDateFormatter pattern)
  (let ((sdf (brl-SimpleDateFormat pattern)))
    (lambda (date)
      ((primitive-constructor <string> (<java.lang.String>))
       ((primitive-virtual-method
	 <java.text.SimpleDateFormat> "format"
	 <java.lang.String>
	 (<java.util.Date>)) sdf date)))))

(define (brl-DecimalFormat pattern)
  ((primitive-constructor <java.text.DecimalFormat>
			  (<String>))
   pattern))

(define (brl-DecimalFormatter pattern)
  (let ((sdf (brl-DecimalFormat pattern)))
    (lambda (num)
      ((primitive-constructor <string> (<java.lang.String>))
       ((primitive-virtual-method
	 <java.text.DecimalFormat> "format"
	 <java.lang.String>
	 (<double>)) sdf num)))))

(define (brl-format obj pattern)
  ((if (number? obj)
       (brl-DecimalFormatter pattern)
       (brl-SimpleDateFormatter pattern))
   obj))

;;;
;;; Various escapes (HTML, LaTeX, etc.)
;;;

(define (brl-string-escaper esc)
  (let ((spec (brl-char-escape-spec esc)))
    (lambda (str) (brl-string-escape str spec))))

(define (brl-string-needs-escape? str esc)
  (let ((len (string-length str)))
    (let loop ((i 0))
      (if (= i len)
	  #f
	  (let ((c (string-ref str i)))
	    (if (and (char>=? c (car esc))
		     (char<=? c (cadr esc)))
		#t
		(loop (+ 1 i))))))))

(define (brl-string-escape str esc)
  (if (brl-string-needs-escape? str esc)
      (list->string
       (reverse
	(let ((len (string-length str)))
	  (let loop ((i 0)
		     (li '()))
	    (if (= i len)
		li
		(loop (+ 1 i)
		      (let ((c (string-ref str i)))
			(if (and (char>=? c (car esc))
				 (char<=? c (cadr esc)))
			    (let ((li2 (vector-ref
					(caddr esc)
					(- (char->integer c)
					   (char->integer (car esc))))))
			      (if li2
				  (append li2 li)
				  (cons c li)))
			    (cons c li)))))))))
      str))

(define (brl-char-escape-spec speclist)
  (let ((minchar (caar speclist))
	(maxchar (caar speclist)))
    (let loop ((li (cdr speclist)))
      (if (not (null? li))
	  (begin
	    (let ((testchar (caar li)))
	      (if (char<? testchar minchar)
		  (set! minchar testchar))
	      (if (char>? testchar maxchar)
		  (set! maxchar testchar)))
	    (loop (cdr li)))))
    (list
     minchar
     maxchar
     (let ((specv (make-vector (+ 1 (- (char->integer maxchar)
				       (char->integer minchar))) #f)))
      (map (lambda (specpair)
	     (vector-set! specv
			  (- (char->integer (car specpair))
			     (char->integer minchar))
			  (reverse (string->list (cdr specpair)))))
	   speclist)
      specv))))

(define brl-html-escape (brl-string-escaper '((#\< . "&lt;")
					      (#\> . "&gt;")
					      (#\" . "&quot;")
					      (#\& . "&amp;"))))

(define brl-scheme-escape (brl-string-escaper '((#\\ . "\\\\")
						(#\" . "\\\""))))

(define brl-sql-escape (brl-string-escaper '((#\' . "''"))))
(define (brl-sql-string obj)
  (if (null? obj)
      "NULL"
      (string-append "'" (brl-sql-escape (brl-string obj)) "'")))

(define brl-latex-escape (brl-string-escaper '((#\\ . "\\\\")
					       (#\~ . "\\~")
					       (#\# . "\\#")
					       (#\$ . "\\$")
					       (#\% . "\\%")
					       (#\^ . "\\^")
					       (#\& . "\\&")
					       (#\{ . "\\{")
					       (#\} . "\\}")
					       (#\_ . "\\_"))))

;;
;; misc
;;

(define (brl-nonblank? obj)
  (not (or (null? obj) (equal? "" obj))))

(define (brl-jclass obj)
   ((primitive-virtual-method <java.lang.Class>
			      "getName" <String> ())
    ((primitive-virtual-method <java.lang.Object>
			       "getClass" <java.lang.Class> ())
     obj)))
